/*
  Devil-Elec
  Programmed and tested with: Arduino IDE v1.8.13
  17.03.2021
*/

#include <avr/interrupt.h>
#include <NanoEveryPin.h>
#include <NanoEveryPinNamen.h>

/*
    Verbinde Pin 2 (Eingang) mit Pin 12 (Ausgang).
    Schließe optional eine Led an Pin 12 an.
    
    Pin D12 erzeugt mittels millis ein Taktsignal.
    Diese Signal geht auf den Interrupt Eingang D2.
    In der ISR von D2 wird ein Flag gesetzt was in der loop ausgewertet
    wird und Led D13 blinken lässt.
*/

InterruptPin<pinPA0> pinA0;   // Pin D2

// für Demo erforderlich
volatile bool pinA0StatusFlag;
OutputPin<12> takt;
OutputPin<13> led;

void setup(void)
{
  Serial.begin(250000);
  Serial.println(F("\nStart #### #### #### ####"));

  /*
      *** verfügbare Methoden ***  
      init();
      enablePullup();
      disablePullup();
      enableInvertIO();
      disableInvertIO();
      isActiv();
      senseDisable();
      senseBothEdges();
      senseRising();
      senseFalling();
      senseLow();
      shutOff();
      deleteFlag();
    
      if ( .isActiv() )
      {
        // ... mach hier sinnvolles
      }
  */

  pinA0.init();
  pinA0.senseBothEdges();  // probiere 'senseRising' und 'senseFalling' aus

  led.init();   // für Demo erforderlich
  takt.init();  // für Demo erforderlich
}

void loop()
{
  taktSignal(1000);
  
  if (pinA0StatusFlag)
  {
    led.toggle();
    pinA0StatusFlag = false;
  }
}

// ************ Funktionen ************

// für Pin D2, D7, 
ISR (PORTA_PORT_vect)
{
  if (pinA0.isActiv() )
  {
    pinA0StatusFlag = true;
    pinA0.deleteFlag();
  }  
}

// für Pin D5, D9, D10
ISR (PORTB_PORT_vect)
{  
}

// für Pin D4, (D0, D1)
ISR (PORTC_PORT_vect)
{  
}

// für D14, D15, D16, D17, D20, D21
ISR (PORTD_PORT_vect)
{  
}

// für Pin D8, D11, D12, D13, 
ISR (PORTE_PORT_vect)
{  
}

// für Pin D3, D6, D18, D19
ISR (PORTF_PORT_vect)
{  
}

void taktSignal (const unsigned long interval)                     
{
  static unsigned long lastMillis {0};
      
  if (millis() - lastMillis >= interval) {
    lastMillis += interval;
    takt.toggle();
  }
}