// Doc_Arduino - german Arduino Forum
// Arduino Nano Every
#pragma once
/*
#if !defined(ARDUINO_ARCH_MEGAAVR) || !defined(ARDUINO_AVR_NANO_EVERY) || !defined(ARDUINO_AVR_ATmega4809)
  #error "This library only support Arduino Nano Every Board with ATmega4809 controller from the megaAVR0 series."
#else
    // both defined
#endif
*/

#if defined(ARDUINO_ARCH_MEGAAVR)
  // is defined
#else
    #error "This library only support controller from the megaAVR0 series. Especially Arduino Nano Every Board with ATmega4809."
#endif

#if defined(ARDUINO_AVR_NANO_EVERY) || defined(ARDUINO_AVR_ATmega4809)
  // is defined
#else
    #error "This library only support Arduino Nano Every Board with ATmega4809 controller from the megaAVR0 series."
#endif



 
#define INLINE inline __attribute__((always_inline))

namespace NanoEvery 
{ 
  namespace Pin 
  { 
    namespace
    {        
      using Register = volatile uint8_t *;
      
      #if defined(__AVR_ATmega4809__)
        #include "include/NanoEveryRegisterComplete.h"  
      #else
        #error "Your board has no ATmega4809 controller!".
      #endif

      // ------------------ VPORTs ------------------      
    
      constexpr Register getDirVPort(const uint8_t pin) {       // Direction    
        return  (Register) addressList[pin].vDir;
      }
      
       constexpr Register getOutVPort(const uint8_t pin) {      // Output Level
        return  (Register) addressList[pin].vOut;
      }
      
       constexpr Register getInVPort(const uint8_t pin) {       // Input Level Status 
        return  (Register) addressList[pin].vIn;
      }
      
       constexpr Register getFlagVPort(const uint8_t pin) {     // Interrupt Flag
        return  (Register) addressList[pin].vIntFlag;
      }
      
      // ------------------ NORMAL PORTs ------------------   
      
      constexpr Register getDirClearPort(const uint8_t pin) {   // Pin Mode is Input
        return  (Register) addressList[pin].dirClear;
      }
      
      constexpr Register getInPort(const uint8_t pin) {         // Input Level Status
        return  (Register) addressList[pin].in;
      }
      
      constexpr Register getDirSetPort(const uint8_t pin) {     // Pin Mode is Output
        return  (Register) addressList[pin].dirSet;
      }
      
      constexpr Register getOutSetPort(const uint8_t pin) {     // Pin Level High
        return (Register) addressList[pin].outSet;
      }
      
      constexpr Register getOutClearPort(const uint8_t pin) {   // Pin Level Low
        return (Register) addressList[pin].outClear;
      }
      
      constexpr Register getOutTogglePort(const uint8_t pin) {  // Pin Level Toggle
        return (Register) addressList[pin].outToggle;
      }
      
      constexpr Register getIntFlagPort(const uint8_t pin) {    // Interrupt Flag
        return (Register) addressList[pin].intFlag;
      }
      
      // ------------------ SPECIALs ------------------ 
      
      constexpr Register getPortCtrlPort(const uint8_t pin) {   // SlewRate
        return (Register) addressList[pin].ctrlPort;
      }
    
      constexpr Register getPinCtrlPort(const uint8_t pin) {    // Pullup, Interrupts, etc.
        return (Register) addressList[pin].ctrlPin;
      }
              
      constexpr uint8_t getMask(const uint8_t pin) {            // Pin Bitmaske
        return  addressList[pin].mask;
      }       
    }

    
    template<uint8_t arduinoPin>
    class Pin 
    {         
      protected:
        
        // Inputs
        void INLINE initInput()
        {     
          *getDirClearPort(arduinoPin) = getMask(arduinoPin);
        }
        
        bool INLINE isOn() const
        {
          return *getInVPort(arduinoPin) & getMask(arduinoPin);
        }
    
        void INLINE enablePullup()
        { 
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | PORT_PULLUPEN_bm;      // Bit 3 Pullup
        }
        
        void INLINE disablePullup()
        { 
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) & ~PORT_PULLUPEN_bm;    
        }
        
        // Outputs
        void INLINE initOutput()
        {
          *getDirSetPort(arduinoPin) = getMask(arduinoPin);
        }
        
        void INLINE setOn()
        {
          *getOutVPort(arduinoPin) = *getOutVPort(arduinoPin) | getMask(arduinoPin);
        }
        
        void INLINE setOff()
        {
          *getOutVPort(arduinoPin) = *getOutVPort(arduinoPin) & ~getMask(arduinoPin);
        }
        
        void INLINE toggle()
        {
          *getInVPort(arduinoPin) = *getInVPort(arduinoPin) | getMask(arduinoPin);
        }
        
        // Specials
        void INLINE enableSlewRate()
        { 
          *getPortCtrlPort(arduinoPin) = *getPortCtrlPort(arduinoPin) | PORT_SRL_bm;        // Bit 0 SlewRate, valid for the complete port
        }
        
        void INLINE disableSlewRate()
        { 
          *getPortCtrlPort(arduinoPin) = *getPortCtrlPort(arduinoPin) & ~PORT_SRL_bm;       
        }
        
        void INLINE enableInvertIO()
        { 
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | PORT_INVEN_bm;       // Bit 7 Inverted I/O
        }
        
        void INLINE disableInvertIO()
        { 
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) & ~PORT_INVEN_bm;       
        }
              
        // Interrupts
        void INLINE reset()
        { 
          *getPinCtrlPort(arduinoPin) = 0;  // *getPinCtrlPort(arduinoPin) & ~PORT_ISC_gm;        // Interrupt disabled but input buffer enabled
        }
        
        void INLINE detectBothEdges()
        { 
          reset();      
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | 0x01;                // Interrupt enabled with sense on both edges
        }
        
        void INLINE detectRising()
        { 
          reset();       
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | 0x02;                // Interrupt enabled with sense on rising edge
        }
        
        void INLINE detectFalling()
        { 
          reset();      
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | 0x03;                // Interrupt enabled with sense on falling edge
        }
        
        void INLINE disable()
        { 
          reset();        
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | 0x04;                // Interrupt and digital input buffer disabled
        }
        
        void INLINE detectLow()
        { 
          reset();        
          *getPinCtrlPort(arduinoPin) = *getPinCtrlPort(arduinoPin) | 0x05;                // Interrupt enabled with sense on low level
        }     

        void INLINE deleteFlag()
        {       
          *getIntFlagPort(arduinoPin) = *getIntFlagPort(arduinoPin) | getMask(arduinoPin); // delete Interrupt Flag
        }   
               
    };

    template<uint8_t arduinoPin>
    class InputPin : protected Pin<arduinoPin> 
    {
      public:      
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        using Pin<arduinoPin>::isOn;
        using Pin<arduinoPin>::enableInvertIO;
        using Pin<arduinoPin>::disableInvertIO;
        
        void  INLINE init()
        {
          Pin<arduinoPin>::initInput();
        }
      
        INLINE operator bool() const
        {
          return  Pin<arduinoPin>::isOn();
        }
    };
    
    template<uint8_t arduinoPin>
    class InterruptPin : protected InputPin<arduinoPin> 
    {
      public:        
        //using Pin<arduinoPin>::reset;
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        using Pin<arduinoPin>::detectBothEdges;
        using Pin<arduinoPin>::detectRising;
        using Pin<arduinoPin>::detectFalling;
        using Pin<arduinoPin>::detectLow;
        using Pin<arduinoPin>::disable;
        using Pin<arduinoPin>::deleteFlag;
        using Pin<arduinoPin>::isOn;
              
        void INLINE init()
        {
          Pin<arduinoPin>::initInput();
        }          
    };
    
    template<uint8_t arduinoPin>
    class TasterGND : protected InputPin<arduinoPin>
    {
      public:
        using InputPin<arduinoPin>::init;
        using Pin<arduinoPin>::enablePullup;
        //using Pin<arduinoPin>::disablePullup;
        //using Pin<arduinoPin>::enableInvertIO;
        //using Pin<arduinoPin>::disableInvertIO;
        
        void  INLINE init()
        {
          Pin<arduinoPin>::enableInvertIO();
          Pin<arduinoPin>::initInput();
        }
        
        bool INLINE pressed() const
        {
          return InputPin<arduinoPin>::isOn();
        }
        
        bool INLINE released() const
        {
          return !InputPin<arduinoPin>::isOn();
        }
        
        INLINE operator bool() const
        {
          return  pressed();
        }
        
        bool INLINE operator ()() const
        {
          return  pressed();
        }
    };
        
    template<uint8_t arduinoPin>
    class OutputPin : protected Pin<arduinoPin>
    {
      public:       
        using Pin<arduinoPin>::setOn;
        using Pin<arduinoPin>::setOff;
        using Pin<arduinoPin>::toggle;
        using Pin<arduinoPin>::isOn;
        using Pin<arduinoPin>::enableSlewRate;
        using Pin<arduinoPin>::disableSlewRate;
        //using Pin<arduinoPin>::enableInvertIO;
        //using Pin<arduinoPin>::disableInvertIO;
      
        void INLINE init()
        {
          Pin<arduinoPin>::initOutput();
        }
       
        void INLINE set(const bool value)
        {
          if(value) setOn(); else setOff();
        }
        
        bool INLINE operator = (const bool value)
        {
          set(value);
          return value;
        }
        
        bool INLINE operator () (const bool value)
        {
          set(value);
          return value;
        }
        
        INLINE operator bool() const
        {
          return Pin<arduinoPin>::isOn();
        }
    };

    template<uint8_t arduinoPin>
    class OutputPinInv : public OutputPin<arduinoPin>
    {
      public:
        void INLINE init()
        {
          // init, ohne Low Puls
          Pin<arduinoPin>::setOn();
          Pin<arduinoPin>::enableInvertIO();
          Pin<arduinoPin>::initOutput();
        }

        INLINE operator bool() const
        {
          return  !Pin<arduinoPin>::isOn();
        }
    };
    
    template<uint8_t arduinoPin>
    class OpenDrainPin : protected Pin<arduinoPin> 
    {
      public:
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        
        void  INLINE init()
        {
          Pin<arduinoPin>::setOff();
          Pin<arduinoPin>::initInput();
        }
        
        void  INLINE enable()             // Ausgang (Low)
        {
          Pin<arduinoPin>::initOutput();
        }
        
        void  INLINE disable()
        {
          Pin<arduinoPin>::initInput();   // Eingang
        }      
    };

  } 
}
