// Doc_Arduino - german Arduino Forum
// 06.01.2012

#pragma once

#if defined(ARDUINO_ARCH_MEGAAVR)
  // is defined
#else
    #error "This library only support controller from the megaAVR0 series. Especially Arduino Nano Every Board with ATmega4809."
#endif

#if defined(ARDUINO_AVR_NANO_EVERY) || defined(ARDUINO_AVR_ATmega4809)
  // is defined
#else
    #error "This library only support Arduino Nano Every Board with ATmega4809 controller from the megaAVR0 series."
#endif

 
#define INLINE inline __attribute__((always_inline))

namespace NanoEvery 
{ 
  namespace Pin 
  { 
    namespace
    {        
      using Register = volatile uint8_t *;
      
      #if defined(__AVR_ATmega4809__)
        #include "include/NanoEveryRegister.h"  
      #else
        #error "Your board has no ATmega4809 controller!".
      #endif

      // ------------------ VPORTs ------------------ //    
      // Direction
      constexpr Register getDirVPort(const uint8_t pin)   { return (Register) (baseAddrList[pin].vport + addrOffset.vDir); }
      
      // Output Level
      constexpr Register getOutVPort(const uint8_t pin)   { return (Register) (baseAddrList[pin].vport + addrOffset.vOut); }
      
      // Input Level Status
      constexpr Register getInVPort(const uint8_t pin)    { return (Register) (baseAddrList[pin].vport + addrOffset.vIn); }
      
      // Interrupt Flag
      constexpr Register getFlagVPort(const uint8_t pin)  { return (Register) (baseAddrList[pin].vport + addrOffset.vIntFlag); }
      
      // ------------------ NORMAL PORTs ------------------ //
      // Pin Mode is Input
      constexpr Register getDirClear(const uint8_t pin)   { return (Register) (baseAddrList[pin].port + addrOffset.dirClear); }
      
      // Input Level Status
      constexpr Register getInput(const uint8_t pin)      { return (Register) (baseAddrList[pin].port + addrOffset.in); }
      
      // Pin Mode is Output
      constexpr Register getDirSet(const uint8_t pin)     { return (Register) (baseAddrList[pin].port + addrOffset.dirSet); }
      
      // Pin Level High
      constexpr Register getOutSet(const uint8_t pin)     { return (Register) (baseAddrList[pin].port + addrOffset.outSet); }
      
      // Pin Level Low
      constexpr Register getOutClear(const uint8_t pin)   { return (Register) (baseAddrList[pin].port + addrOffset.outClear); }
      
      // Pin Level Toggle
      constexpr Register getOutToggle(const uint8_t pin)  { return (Register) (baseAddrList[pin].port + addrOffset.outToggle); }
      
      // Interrupt Flag
      constexpr Register getIntFlag(const uint8_t pin)    { return (Register) (baseAddrList[pin].port + addrOffset.intFlag); }
      
      // ------------------ SPECIALs ------------------ //
      // SlewRate
      constexpr Register getPortCtrl(const uint8_t pin)   { return (Register) (baseAddrList[pin].port + addrOffset.portCtrl); }
      
      // Invert, Input-Level, Pullup, Interrupt
      constexpr Register getPinNctrl(const uint8_t pin)   { return (Register) (baseAddrList[pin].port + baseAddrList[pin].nCtrl); }
      
      // Pin Bitmaske
      constexpr uint8_t getMask(const uint8_t pin)        { return (baseAddrList[pin].mask); } 
    }

    
    template<uint8_t arduinoPin>
    class Pin 
    {         
      protected:
        
        // Inputs
        void INLINE initInput() {
          *getDirClear(arduinoPin) = getMask(arduinoPin);
        }
        
        bool INLINE isOn() const {
          return *getInVPort(arduinoPin) & getMask(arduinoPin);
        }
    
        void INLINE enablePullup() { 
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | PORT_PULLUPEN_bm;      // Bit 3 Pullup
        }
        
        void INLINE disablePullup() { 
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) & ~PORT_PULLUPEN_bm;    
        }
                   
        // Outputs
        void INLINE initOutput() {
          *getDirSet(arduinoPin) = getMask(arduinoPin);
        }
        
        void INLINE setOn() {
          *getOutVPort(arduinoPin) = *getOutVPort(arduinoPin) | getMask(arduinoPin);
        }
        
        void INLINE setOff() {
          *getOutVPort(arduinoPin) = *getOutVPort(arduinoPin) & ~getMask(arduinoPin);
        }
        
        void INLINE toggle() {
          *getInVPort(arduinoPin) = *getInVPort(arduinoPin) | getMask(arduinoPin);
        }
        
        // Specials
        void INLINE enableSlewRate() { 
          *getPortCtrl(arduinoPin) = *getPortCtrl(arduinoPin) | PORT_SRL_bm;        // Bit 0 SlewRate, valid for the complete port
        }
        
        void INLINE disableSlewRate() { 
          *getPortCtrl(arduinoPin) = *getPortCtrl(arduinoPin) & ~PORT_SRL_bm;       
        }
        
        void INLINE enableInvertIO() { 
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | PORT_INVEN_bm;       // Bit 7 Inverted I/O
        }
        
        void INLINE disableInvertIO() { 
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) & ~PORT_INVEN_bm;       
        }
              
        // Interrupts
        void INLINE reset() { 
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) & ~PORT_ISC_gm;        // all Interrupts disabled but input buffer enabled
        }
        
        void INLINE senseBothEdges() { 
          reset();      
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | 0x01;                // Interrupt enabled with sense on both edges
        }
        
        void INLINE senseRising() { 
          reset();       
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | 0x02;                // Interrupt enabled with sense on rising edge
        }
        
        void INLINE senseFalling() { 
          reset();      
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | 0x03;                // Interrupt enabled with sense on falling edge
        }
        
        void INLINE senseDisable() { 
          reset();        
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | 0x04;                // Interrupt and digital input buffer disabled
        }
        
        void INLINE senseLow() { 
          reset();        
          *getPinNctrl(arduinoPin) = *getPinNctrl(arduinoPin) | 0x05;                // Interrupt enabled with sense on low level
        }     

        void INLINE deleteFlag() {       
          *getFlagVPort(arduinoPin) = getMask(arduinoPin);                           // delete Interrupt Flag
        }
    };

    template<uint8_t arduinoPin>
    class InputPin : protected Pin<arduinoPin> 
    {
      public:      
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        using Pin<arduinoPin>::isOn;
        using Pin<arduinoPin>::enableInvertIO;
        using Pin<arduinoPin>::disableInvertIO;
        
        void  INLINE init()
        {
          Pin<arduinoPin>::initInput();
        }
      
        INLINE operator bool() const
        {
          return  Pin<arduinoPin>::isOn();
        }
    };
    
    template<uint8_t arduinoPin>
    class InterruptPin : protected InputPin<arduinoPin> 
    {
      public:        
        //using Pin<arduinoPin>::reset;
        using Pin<arduinoPin>::enablePullup;
        using Pin<arduinoPin>::disablePullup;
        using Pin<arduinoPin>::senseBothEdges;
        using Pin<arduinoPin>::senseRising;
        using Pin<arduinoPin>::senseFalling;
        using Pin<arduinoPin>::senseLow;
        using Pin<arduinoPin>::senseDisable;
        using Pin<arduinoPin>::deleteFlag;
        using Pin<arduinoPin>::isOn;
              
        void INLINE init()
        {
          Pin<arduinoPin>::initInput();
        }          
    };
    
    template<uint8_t arduinoPin>
    class TasterGND : protected InputPin<arduinoPin>
    {
      public:
        using InputPin<arduinoPin>::init;
        using Pin<arduinoPin>::enablePullup;
        //using Pin<arduinoPin>::disablePullup;
        //using Pin<arduinoPin>::enableInvertIO;
        //using Pin<arduinoPin>::disableInvertIO;
        
        void  INLINE init() {
          Pin<arduinoPin>::enableInvertIO();
          Pin<arduinoPin>::initInput();
        }
        
        bool INLINE pressed() const     { return InputPin<arduinoPin>::isOn(); }
        
        bool INLINE released() const    { return !InputPin<arduinoPin>::isOn(); }
        
        INLINE operator bool() const    { return pressed(); }
        
        bool INLINE operator ()() const { return pressed(); }
    };
        
    template<uint8_t arduinoPin>
    class OutputPin : protected Pin<arduinoPin>
    {
      public:       
        using Pin<arduinoPin>::setOn;
        using Pin<arduinoPin>::setOff;
        using Pin<arduinoPin>::toggle;
        using Pin<arduinoPin>::isOn;
        using Pin<arduinoPin>::enableSlewRate;
        using Pin<arduinoPin>::disableSlewRate;
        using Pin<arduinoPin>::enableInvertIO;
        using Pin<arduinoPin>::disableInvertIO;
      
        void INLINE init() { Pin<arduinoPin>::initOutput(); }
       
        void INLINE set(const bool value) {
          if(value) setOn();
          else setOff();
        }
        
        bool INLINE operator = (const bool value) {
          set(value);
          return value;
        }
        
        bool INLINE operator () (const bool value) {
          set(value);
          return value;
        }
        
        INLINE operator bool() const { return Pin<arduinoPin>::isOn(); }
    };

    template<uint8_t arduinoPin>
    class OutputPinInvert : public OutputPin<arduinoPin>
    {
      public:
        void INLINE init()
        {
          // init, ohne Low Puls
          Pin<arduinoPin>::setOn();
          Pin<arduinoPin>::enableInvertIO();
          Pin<arduinoPin>::initOutput();
        }

        INLINE operator bool() const { return !Pin<arduinoPin>::isOn(); }
    };

  } 
}
