// Doc_Arduino - german Arduino Forum
// 07.02.2021

#pragma once


#if defined(ARDUINO_AVR_NANO_EVERY) && defined(__AVR_ATmega4809__)
    // is defined
    #include <NanoEveryPin.h>
#else
    #error "This library only support Arduino Nano Every Board with ATmega4809."
#endif
 
#define INLINE inline __attribute__((always_inline))

using namespace NanoEvery::Pin;
using Register = volatile uint8_t *;

namespace Intern
{
  // Outputs
  void INLINE output(const uint8_t arduinoPin) {
    *regVPORTdir(arduinoPin) = *regVPORTdir(arduinoPin) | getMask(arduinoPin);
  }

  void INLINE on(const uint8_t arduinoPin) {
    *regVPORTout(arduinoPin) = *regVPORTout(arduinoPin) | getMask(arduinoPin);
  }

  void INLINE off(const uint8_t arduinoPin) {
    *regVPORTout(arduinoPin) = *regVPORTout(arduinoPin) & ~getMask(arduinoPin);
  }
}        

// for Debugging and Tests etc.

// ------------------ VPORTs ----------------------------------------------------------------------------------------------- //                
constexpr Register getRegVPORTdir(const uint8_t pin)       { return (Register) regVPORTdir(pin); }
constexpr Register getRegVPORTdirPORTA(const uint8_t pin)  { return (Register) regVPORTdirPORTA(pin); }
constexpr Register getRegVPORTout(const uint8_t pin)       { return (Register) regVPORTout(pin); }
constexpr Register getRegVPORTin(const uint8_t pin)        { return (Register) regVPORTin(pin); }
constexpr Register getRegVPORTflag(const uint8_t pin)      { return (Register) regVPORTflag(pin); }

// ------------------ NORMAL PORTs ----------------------------------------------------------------------------------------- //
constexpr Register getRegPORTdirClear(const uint8_t pin)   { return (Register) regPORTdirClear(pin); }
constexpr Register getRegPORTin(const uint8_t pin)         { return (Register) regPORTin(pin); }
constexpr Register getRegPORTdirSet(const uint8_t pin)     { return (Register) regPORTdirSet(pin); }
constexpr Register getRegPORToutSet(const uint8_t pin)     { return (Register) regPORToutSet(pin); }
constexpr Register getRegPORToutClear(const uint8_t pin)   { return (Register) regPORToutClear(pin); }
constexpr Register getRegPORToutToggle(const uint8_t pin)  { return (Register) regPORToutToggle(pin); }
constexpr Register getRegPORTflag(const uint8_t pin)       { return (Register) regPORTflag(pin); }

// ------------------ SPECIALs ------------------ //
constexpr Register getRegPortCtrl(const uint8_t pin)       { return (Register) regPortCtrl(pin); }
constexpr Register getRegPinCtrl(const uint8_t pin)        { return (Register) regPinCtrl(pin); }
constexpr Register getRegPinCtrlPORTA(const uint8_t pin)   { return (Register) regPinCtrlPORTA(pin); }
  
// funktioniert: constexpr uint8_t contentdir(const uint8_t p) { return *(uint8_t*) (getRegVPORTdir(p)); }
// funktioniert: constexpr uint8_t contentdir(const uint8_t p) { return *(const_cast<uint8_t*>(getRegVPORTdir(p))); }
// besser:
// gibt den jeweiligen Registerinhalt zurück
constexpr uint8_t contentDIR(const uint8_t p)       { return *(const_cast<uint8_t*>(regVPORTdir(p))); }
constexpr uint8_t contentIN(const uint8_t p)        { return *(const_cast<uint8_t*>(regVPORTin(p)) ); }
constexpr uint8_t contentOUT(const uint8_t p)       { return *(const_cast<uint8_t*>(regVPORTout(p))); }
constexpr uint8_t contentPINCTRL(const uint8_t p)   { return *(const_cast<uint8_t*>(regPinCtrl(p)) ); }
constexpr uint8_t contentPORTCTRL(const uint8_t p)  { return *(const_cast<uint8_t*>(regPortCtrl(p))); }

// gibt die jeweilige Registeradresse zurück
constexpr uint16_t addrDIR(const uint8_t p)       { return reinterpret_cast<uint16_t>(regVPORTdir(p)); }
constexpr uint16_t addrIN(const uint8_t p)        { return reinterpret_cast<uint16_t>(regVPORTin(p) ); }
constexpr uint16_t addrOUT(const uint8_t p)       { return reinterpret_cast<uint16_t>(regVPORTout(p)); }
constexpr uint16_t addrPINCTRL(const uint8_t p)   { return reinterpret_cast<uint16_t>(regPinCtrl(p) ); }
constexpr uint16_t addrPORTCTRL(const uint8_t p)  { return reinterpret_cast<uint16_t>(regPortCtrl(p)); }

void pinInitIntern (const uint8_t pin)
{
  Intern::output(pin);
}

enum class reg : uint8_t {DIR, IN, OUT, PIN, PORT};  // Registerauswahl
enum class pinRow : uint8_t {LOWER, UPPER, ALL};     // lower D0...D12, upper D13...D21

uint32_t getBinary (const reg what)
{
  uint32_t content {0};
  uint32_t bitSchubser {1};
  uint8_t portBit {0};
  
  for (uint8_t i = 0; i < ANZAHLPINS; i++) {
    switch(what) {
      case reg::DIR:  portBit = contentDIR(i);      break;
      case reg::IN:   portBit = contentIN(i);       break;
      case reg::OUT:  portBit = contentOUT(i);      break;
      case reg::PIN:  portBit = contentPINCTRL(i);  break;
      case reg::PORT: portBit = contentPORTCTRL(i); break;
    }
    
    if (portBit & getMask(i) ) {
      content |= bitSchubser;
    }
    bitSchubser <<= 1;
  }
  return content;
}

uint32_t getBinary (const reg what, const pinRow row)
{
  uint32_t content {0};
  uint32_t bitPinPos {1};
  uint8_t portBit {0};
  
  uint8_t firstPin {0};
  uint8_t lastPin {ANZAHLPINS-1};
 
  switch(row) {
    case pinRow::LOWER: lastPin = 12;   break;  // D0....D12
    case pinRow::UPPER: firstPin = 13;  break;  // D13...D21
    case pinRow::ALL:                   break;  // D0....D21
  }
  
  for (uint8_t i = firstPin; i <= lastPin; i++)
  { // Inhalt vom gewünschten Pin-Register lesen
    switch(what) {
      case reg::DIR:  portBit = contentDIR(i);      break;
      case reg::IN:   portBit = contentIN(i);       break;
      case reg::OUT:  portBit = contentOUT(i);      break;
      case reg::PIN:  portBit = contentPINCTRL(i);  break;
      case reg::PORT: portBit = contentPORTCTRL(i); break;
    }
    // Bitmaske vom Pinnummer 'i' holen und 
    if (portBit & getMask(i) )
    { // an durchnummerierte Pin-Position schreiben
      content |= bitPinPos;   
    }
    bitPinPos <<= 1;    // für nächsten Durchlauf Pin-Position um eins weiterschieben
  }
  
  return content;
}
// 3317 / 177 zu 3357 / 177
template <class T>
void formattingBinary (Stream &out, const T &var, const pinRow row)
{
  //uint32_t mask {1};   // wegen Overflow Compilerwarnung aufgetrennt
  //mask = (mask << (anz-1));
  
  uint32_t mask {0};
  uint8_t stellen {22};
  switch(row) {
    case pinRow::LOWER: stellen = 13; mask = 0x1000;   break;  // D0....D12
    case pinRow::UPPER: stellen =  9; mask = 0x0100;   break;  // D13...D21
    case pinRow::ALL:                 mask = 0x200000; break;  // D0....D21
  }
/*  
  // switch kostet 40 Bytes mehr Flash, dafür konstant niedrige 1,3µs
  uint32_t mask {0};
  switch(anz) {
    case  9: mask = 0x0100;   break;  // upper, D13...D21
    case 13: mask = 0x1000;   break;  // lower, D0....D12
    case 22: mask = 0x200000; break;  // all,   D0....D21
  }
*/
  uint8_t digit {0};
  //out.print("0b");
  for (uint8_t i = 0; i < stellen; i++)
  {
    if (var & mask) digit = 1;
    else digit = 0;
    out.print(digit);
/*
    // aller 4 Bits ein Punkt dazwischen schieben und das Letzte unterdrücken
    if (!((i+1)%4) && (i<(anz-1))) {
      out.print(".");
    }
*/    
    mask = mask >> 1;
  }
}
// 3323 / 177 Bytes
void showBinaryln (Stream &out, const reg what, const pinRow row)
{
  Intern::on(2); 
  /*
  uint8_t stellen {22};

  switch(row) {
    case pinRow::LOWER: stellen = 13;  break;  // D0....D12
    case pinRow::UPPER: stellen =  9;  break;  // D13...D21
    case pinRow::ALL:                  break;  // D0....D21
  }
  */  
  formattingBinary(out, getBinary(what, row), row);
  Intern::off(2); 
  out.println();
}
    
    /*
uint32_t getBinaryOld (const reg what, const pinRow row)
{
  Intern::on(2);
  uint32_t content {0};
  uint32_t bitSchubser {1};
  uint8_t portBit {0};
  
  for (uint8_t i = 0; i < ANZAHLPINS; i++)
  { // Inhalt vom gewünschten Pin-Register lesen
    switch(what) {
      case reg::DIR:  portBit = contentDIR(i);      break;
      case reg::IN:   portBit = contentIN(i);       break;
      case reg::OUT:  portBit = contentOUT(i);      break;
      case reg::PIN:  portBit = contentPINCTRL(i);  break;
      case reg::PORT: portBit = contentPORTCTRL(i); break;
    }
    // Bitmaske vom Pin 'i' holen und 
    if (portBit & getMask(i) )
    { // an durchnummerierte Pin-Position schreiben
      content |= bitSchubser;   
    }
    bitSchubser <<= 1;    // durchnummerierte Pin-Position um eins weiterschieben
  }
  
  // jetzt steht für jeden Arduino Pin eine 0 oder 1 an entsprechender Position in der Variablen 'content'
  
  // hier wird nur noch je nach äußere Board Pinreihe aufgeteilt und zurechtgeschoben
  switch(row) {
    case pinRow::LOWER:                 break;  // D12....D0
    case pinRow::UPPER: content >>= 13; break;  // D21...D13
    case pinRow::ALL:                   break;  // D21....D0
  }
  Intern::off(2);
  return content;
}
*/


uint16_t getBinaryLower (const reg what)
{
  uint32_t bitContent { getBinary (what) };
  bitContent &= 0x1FFF;
  return static_cast<uint16_t>(bitContent);
}

uint16_t getBinaryUpper (const reg what)
{
  uint32_t bitContent { getBinary (what) };
  bitContent >>= 13;
  bitContent &= 0x01FF;
  return static_cast<uint16_t>(bitContent);
}
