/*
  Doc_Arduino - german Arduino Forum
  Programmed and tested with: Arduino IDE 1.8.13 & avr-gcc 10.2.0
  25.02.2021
  License: GNU GPLv3
*/

#pragma once

#include <NanoEveryPin.h>

// zum Lib internen Zeiten messen,
// dann hier irgendwo bei Befarf Intern::on(x); bzw. Intern::off(x); verwenden und
// messPinIntern(x) in setup aufrufen
namespace Intern
{
  // Outputs
  void INLINE output(const uint8_t arduinoPin) {
    *regVPORTdir(arduinoPin) = *regVPORTdir(arduinoPin) | getMask(arduinoPin);
  }

  void INLINE on(const uint8_t arduinoPin) {
    *regVPORTout(arduinoPin) = *regVPORTout(arduinoPin) | getMask(arduinoPin);
  }

  void INLINE off(const uint8_t arduinoPin) {
    *regVPORTout(arduinoPin) = *regVPORTout(arduinoPin) & ~getMask(arduinoPin);
  }
}        

void messPinIntern (const uint8_t pin)
{
  Intern::output(pin);
}

// ********* for debugging or what ever ********* //
// returns the respective register content
constexpr uint8_t contentDIR(const uint8_t p)       { return *(const_cast<uint8_t*>(regVPORTdir(p))); }
constexpr uint8_t contentIN(const uint8_t p)        { return *(const_cast<uint8_t*>(regVPORTin(p)) ); }
constexpr uint8_t contentOUT(const uint8_t p)       { return *(const_cast<uint8_t*>(regVPORTout(p))); }
constexpr uint8_t contentPINCTRL(const uint8_t p)   { return *(const_cast<uint8_t*>(regPinCtrl(p)) ); }
constexpr uint8_t contentPORTCTRL(const uint8_t p)  { return *(const_cast<uint8_t*>(regPortCtrl(p))); }

// returns the respective register address
constexpr uint16_t addrDIR(const uint8_t p)       { return reinterpret_cast<uint16_t>(regVPORTdir(p)); }
constexpr uint16_t addrIN(const uint8_t p)        { return reinterpret_cast<uint16_t>(regVPORTin(p) ); }
constexpr uint16_t addrOUT(const uint8_t p)       { return reinterpret_cast<uint16_t>(regVPORTout(p)); }
constexpr uint16_t addrPINCTRL(const uint8_t p)   { return reinterpret_cast<uint16_t>(regPinCtrl(p) ); }
constexpr uint16_t addrPORTCTRL(const uint8_t p)  { return reinterpret_cast<uint16_t>(regPortCtrl(p)); }


enum class reg : uint8_t {IN, OUT, DIR, PIN, PORT};   // Register selection
enum class row : uint8_t {LOWER, UPPER, ALL};         // Board pin row, PCB left/right||top/down >> lower D0...D12, upper D13...D21

namespace BinaryFormatter
{
    // Liest für die gewünschte PCB Pinreihe das gewünschte Register aus,
    // setzt die Bits entsprechend der Arduino Pinnummer Reihenfolge zusammen
    // und gibt den Wert zurück.
    uint32_t getBinary (const reg what, const row pinRow)
    {
      uint32_t content {0};
      uint32_t bitPositionShifter {1};
      uint8_t portBit {0};
      
      uint8_t firstPin {0};
      uint8_t lastPin {Addr::ANZAHLPINS-1};      // NanoEveryRegister.h
     
      switch(pinRow) {
        case row::LOWER: lastPin = 12;   break;  // D0....D12
        case row::UPPER: firstPin = 13;  break;  // D13...D21
        case row::ALL:                   break;  // D0....D21
      }
      
      // Inhalt vom gewünschten Pin-Register lesen
      for (uint8_t i = firstPin; i <= lastPin; i++) {
        switch(what) {
          case reg::IN:   portBit = contentIN(i);       break;
          case reg::OUT:  portBit = contentOUT(i);      break;
          case reg::DIR:  portBit = contentDIR(i);      break;
          case reg::PIN:  portBit = contentPINCTRL(i);  break;
          case reg::PORT: portBit = contentPORTCTRL(i); break;
        }
        // Bitmaske vom Pinnummer 'i' holen und 
        if (portBit & getMask(i) )
        { // an durchnummerierte Pin-Position schreiben
          content |= bitPositionShifter;   
        }
        bitPositionShifter <<= 1;   // für nächsten Durchlauf Pin-Position um eins weiterschieben
      }
      return content;
    }

    // Der Rückgabewert von getBinary wird bei der Ausgabe entsprechend auf die Stellen der PCB Pinreihe begrenzt.
    // Damit wird verhindert das unnötige Nullen vom Datentyp bedingt ausgegeben werden.
    template<class T>
    void formatBinary (Stream &out, const T &var, const row pinRow)
    {  
      uint32_t mask {0};
      uint8_t stellen {22};
      switch(pinRow) {
        case row::LOWER: stellen = 13; mask = 0x00001000; break;  // D0....D12
        case row::UPPER: stellen =  9; mask = 0x00000100; break;  // D13...D21
        case row::ALL:                 mask = 0x00200000; break;  // D0....D21
      }

      uint8_t digit {0};
      //out.print("0b");
      for (uint8_t i = 0; i < stellen; i++) {
        if (var & mask) digit = 1;
        else digit = 0;
        out.print(digit);
       
        mask = mask >> 1;
      }
    }
}

// Aufrufer Funktionen im Sketch mit oder ohne Linefeed.
void showBinary (Stream &out, const reg what, const row pinRow)
{
  using namespace BinaryFormatter;
  formatBinary(out, getBinary(what, pinRow), pinRow);
}

void showBinaryln (Stream &out, const reg what, const row pinRow)
{
  using namespace BinaryFormatter;
  formatBinary(out, getBinary(what, pinRow), pinRow);
  out.println();
}
